;Baustelle

/*
Einfaches Voltmeter mit Auto-Range fr AC und DC mit Tiny24, TL072, LCD 10x1 
oder LCD 8x1 und 2 Kleinrelais zur Messbereichsumschaltung.
Die Messbereiche sind binr gestuft, um die Skalierung einfach zu halten.

Exemplarstreuungen der internen Referenzspannung lassen sich durch separate
Kalibrierwerte fr alle Messbereiche, sowie AC- und DC-Anzeigewert 
kompensieren. Dies erfolgt mittels eines separaten Projektes, dessen
Quelltext frei verfgbar ist und das die Datei "VM24l_kalibration.eep"
erzeugt, die zum Kalibrieren ins EEPROM geladen werden muss.

*/

#define servicemode         ;aktiviert Service-Mode mit LCD 16x2 
;Der Service-Mode dient zum berprfen der Analogschaltung. Dazu werden auf
;einem LCD 16x2 einige Zwischenwerte angezeigt, die im finalen Gert nicht
;verfgbar sind. Dies wren: Messwert Plus-Seite, Messwert Minus-Seite und
;Messbereich. Und natrlich (wie im Endgert) Messwertsumme und Polaritt.

;#define lcd8x1              ;Verwendung eines LCDs mit 1 Zeile je 8 Zeichen

;----------------------------------------------------------------------------
.include "tn24def.inc"  ;Portnamen
.equ clock=1000000      ;Prozessortakt 1 MHz
.equ acz=7              ;Zeichen fr AC

.equ limu=40            ;Limit unten, Schwelle zum Messbereich vermindern
.equ limc=3             ;Limit unten zur C-Text-Ausgabe
.equ ctw0=127           ;Wartezeit bis Scrolltext zuschlgt
.equ anz0u=10           ;erstes Anzeige-Intervall nach Messbereichsumschaltung
.equ anz0=20            ;Anzeige-Intervall im Normalbetrieb
.equ anzs=20            ;Anzeige-Intervall beim Copyright-Scroll-Text
.equ relwait=2000       ;Wartezeit fr Relais-Reaktion in 11s
.equ maxdiff=50         ;maximal erlaubte Differenz fr Wechselspannung


#ifdef servicemode  ;LCD 16x2 zur Inbetriebnahme
.equ zln=2              ;LCD 16x2 ist als
.equ spn=16             ;16x2 organisiert
#elif lcd8x1        ;LCD 8x1
.equ zln=2              ;LCD 8x1 ist als
.equ spn=4              ;4x2 organisiert
#else               ;LCD 10x1 fr finales Gert
.equ zln=2              ;LCD 10x1 ist als
.equ spn=5              ;5x2 organisiert
#endif 

.equ mqpos=(1<<refs1)+7 ;Messquelle fr positiven Messwert (ADMUX)
.equ mqneg=(1<<refs1)+6 ;Messquelle fr negativen Messwert (ADMUX)
.equ kalid=4            ;EEP-Adresse fr Kalibrierwert DC
.equ kalia=5            ;EEP-Adresse fr Kalibrierwert AC


.dseg                   ;Variablen im SRAM
.org sram_start         ;ab Beginn des SRAMs
;LCD
;vlcd:           .byte zln*spn   ;virt. Bildschirmspeicher
vlcd:           .byte 34    ;vollen Platz fr LCD 16x2 reservieren
rpos:           .byte 1     ;Leseposition VLCD
wpos:           .byte 1     ;Schreibposition VLCD
blinka:         .byte 1     ;Blinkbereich Anfang
blinke:         .byte 1     ;Blinkbereich Ende
;Print:
tmpmath:        .byte 4     ;Datenbergabebereich fr Zahl -> ASCII
cindex:         .byte 1     ;Scrolltext-Index
signum:         .byte 1     ;Vorzeichen als Textzeichen
maxwert:        .byte 1     ;Maximalwert (Spitzenwert) einer Messperiode
;ADC
atab:           .byte 16    ;je 4x ADCWH, ADCWL, ADCWn, Minwert
    .equ dc=0               ;Index kalibrierter DC-Wert
    .equ pl=1               ;Index Wert am Plus-Eingang
    .equ ac=2               ;Index kalibrierter AC-Wert
    .equ mi=3               ;Index Wert am Minus-Eingang
    .equ mwh=0              ;Offset Mittelwert H-Byte
    .equ mwl=4              ;Offset Mittelwert L-Byte
    .equ mwn=8              ;Offset Mittelwert Nachkomma-Byte
    .equ min=12             ;Offset Minimalwert
    .equ mux=16             ;Offset Wert fr ADMUX
muxtab:         .byte 4     ;ADMUX-Werte der Mess-Stellen
reltab:         .byte 4     ;Relais-Bitmuster fr 4 Messbereiche


;Variablen in Registern:
.def suml=r1                ;unscalierte Summe der Messwerte L
.def sumh=r2                ;unscalierte Summe der Messwerte H
.def srsk=r3                ;SREG-Sicherung
.def bliz=r4                ;Blinkzhler LCD
.def kaz=r5                 ;Kanal-Zhler und Anzeigezeit-Vorteiler
.def it0=r6                 ;ungesichert
.def it1=r7                 ;in der
.def it2=r8                 ;ADC-ISR
.def it3=r9                 ;zum Rechnen
.def mw0=r10                ;Messwert, Byte 0 (Nachkommastellen)
.def mw1=r11                ;Messwert, Byte 1 (LSB)
.def mw2=r12                ;Messwert, Byte 2
.def mw3=r13                ;Messwert, Byte 3 (MSB)


.def ctwait=r18             ;Wartezeitzhler fr Copyright-Scrolltext
.def mb=r19                 ;Messbereich (Index)
.def itmp=r20               ;temporr in ISR
.def anz=r21                ;Anzeigetakt-Zhler
.def vflags=r22             ;Flags fr LCD und Anderes
    .equ lsu=7                  ;letzte Stelle unterdrcken
    .equ mbu=6                  ;es fand eine Messbereichsumschaltung statt
    .equ zeiwe=5                ;Flag fr Zeilenwechsel
    .equ update=4               ;Flag fr Update LCD
    .equ trennbit=3             ;Trennbit gegen berlauf
    .equ upsync=2               ;Synchronisationsflag vom Timer-Interrupt
    ;Bit 0 und 1 dienen als Zhler fr upsync

.def tmp=r23                ;temporr (ungesichert in LCD-Routinen)
.def wl=r24                 ;temp l
.def wh=r25                 ;temp h
;.def null=r27               ;yh ist immer 0

;----------------------------------------------------------------------------

.cseg                   ;Programm-Speicher
.org 0                  ;ab Adresse 0
 rjmp RESET         ;Reset Handler
 rjmp isr_exi0      ;EXT_INT0      ;IRQ0 Handler
 rjmp nix;PCINT0        ;PCINT0 Handler
 rjmp nix;PCINT1        ;PCINT1 Handler
 rjmp nix;WATCHDOG      ;Watchdog Interrupt Handler
 rjmp nix;TIM1_CAPT     ;Timer1 Capture Handler
 rjmp nix;TIM1_COMPA    ;Timer1 Compare A Handler
 rjmp nix;TIM1_COMPB    ;Timer1 Compare B Handler
 rjmp nix;TIM1_OVF      ;Timer1 Overflow Handler
 rjmp nix;TIM0_COMPA    ;Timer0 Compare A Handler
 rjmp nix;TIM0_COMPB    ;Timer0 Compare B Handler
 rjmp nix;TIM0_OVF      ;Timer0 Overflow Handler
 rjmp nix;ANA_COMP      ;Analog Comparator Handler
 rjmp isr_adcc      ;ADCC          ;ADC Conversion Handler
 rjmp nix;EE_RDY        ;EEPROM Ready Handler
 rjmp nix;USI_STR       ;USI STart Handler
 rjmp nix;USI_OVF       ;USI Overflow Handler
nix:
 rjmp nix

.include "LCD_10x1.inc"     ;LCD-Routinen
.include "pr_i2a16.inc"     ;Ausgabe-Routine fr 16-Bit-Zahl zu ASCII
.include "ctext.inc"        ;Copyright-Text
.include "pr_i2a24.inc"     ;Ausgabe-Erweiterung auf 24-Bit-Zahl
.include "LCD_Udef.inc"     ;benutzerdefinierte Zeichen
;.include "pr_debug.inc"     ;Ausgabe-Erweiterung Hex und Bin



;----------------------------------------------------------------------------

isr_adcc:           ;ISR, ADC Complete, neues Messergebnis liegt vor
 push yl
 in srsk,sreg               ;PSW sichern
 ldi yl,atab                ;Y-Pointer
 mov itmp,kaz               ;auf 
 andi itmp,3                ;Nummer der
 add yl,itmp                ;Mess-Stelle
 ldd itmp,y+mux             ;nchste Mess-Stelle
 out admux,itmp             ;auswhlen
 ldd it0,y+mwn              ;ADC-Nachkommawert,
 ldd it1,y+mwl              ;ADC-Mittelwert L und
 ldd it2,y+mwh              ;ADC-Mittelwert H holen
 in it3,adcl                ;ADC-Messwert
 in itmp,adch               ;auslesen
 sbrs vflags,mbu            ;Messbereichsumschaltung? - ja...
 rjmp adcc_miwert           ;nein, Mittelwert bilden...

 mov it0,yh                 ;ja, Messwert 
 mov it1,it3                ;direkt (ohne Mittelwertbildung)
 mov it2,itmp               ;bernehmen
 rjmp adcc_save             ;weiter zum Sichern der Werte

adcc_miwert:
 sub it0,it1                ;1/256 des
 sbc it1,it2                ;Mittelwert-Inhaltes
 sbc it2,yh                 ;herausnehmen
 add it0,it3                ;1/256 des neuen Messwertes
 adc it1,itmp               ;zum Mittelwert-Inhalt
 adc it2,yh                 ;dazulegen
 lsr itmp                   ;ADC-Wert
 ror it3                    ;von 10 Bit Auflsung
 lsr itmp                   ;auf 8 Bit Auflsung
 ror it3                    ;reduzieren,

 ldd itmp,y+min             ;bisherigen Minimalwert holen
 cp it3,itmp                ;ist aktueller Wert kleiner?
 brsh adcc_mawert           ;nein...
 sbrc kaz,0                 ;war das die jeweils zweite Messung? - nein...
 std y+min,it3              ;ja, neuen (kleineren) Minimalwert sichern
adcc_mawert:
 lds itmp,maxwert           ;bisherigen Spitzenwert holen
 cp it3,itmp                ;ist aktueller Wert grer?
 brlo adcc_save             ;nein...
 sbrc kaz,0                 ;war das die jeweils zweite Messung? - nein...
 sts maxwert,it3            ;ja, neuen (greren) Spitzenwert sichern

adcc_save:
 sbrs kaz,0                 ;war das die jeweils erste Messung? - nein...
 rjmp adcc_save1            ;ja, Messwert verwerfen...
 std y+mwn,it0              ;Mittelwert wieder
 std y+mwl,it1              ;im SRAM
 std y+mwh,it2              ;sichern
adcc_save1:
 inc kaz                    ;Kanal-Zhler hoch
 brne adcc_save2            ;berlauf? - nein...
 dec anz                    ;ja, Anzeigezhler runter
adcc_save2:
 sbrc kaz,7                 ;nach 128 Messungen (32 Runden) die
 cbr vflags,1<<mbu          ;Messbereichsumschaltung entwerten
 inc vflags                 ;Zhler fr LCD-Ausgabe erhhen,
 cbr vflags,1<<trennbit     ;dabei berlauf verhindern
 out sreg,srsk              ;PSW wiederherstellen
 pop yl
 reti                       ;fertig, zurck...

isr_exi0:           ;ISR, berspannung an einem Messeingang
 in srsk,sreg               ;PSW sichern
 inc mb                     ;Messbereich hochschalten
 cpi mb,4                   ;und
 brlo exi0_1                ;auf 3
 ldi mb,3                   ;begrenzen
exi0_1:
 rcall sub_neumb            ;Relais gem Messbereich umschalten...
 out sreg,srsk              ;PSW wiederherstellen
 reti                       ;fertig, zurck...



;----------------------------------------------------------------------------

RESET:
 ldi wl,low(RAMEND)         ;Stackpointer-Init
 out SPL,wl
 clr yh                     ;immer RAM-Seite 0 als Y-Pointer und immer 0
 ;Fixwerte fr ADMUX und Relaisbitmuster zum schnellen Zugriff ber den
 ;Messzyklus-Index ins SRAM kopieren:
 ldi zl,low(f_muxtab*2)     ;Z-Pointer auf
 ldi zh,high(f_muxtab*2)    ;Array mit Fixwerten im Flash
 ldi yl,muxtab              ;Array mit Fixwerten im SRAM
reset1:
 lpm r0,z+                  ;Fixwerte fr ADMUX und Relais
 st y+,r0                   ;aus Flash ins SRAM kopieren
 cpi yl,muxtab+8            ;alle 8 Werte fertig?
 brne reset1                ;nein, nochmal...
 ;I/O initialisieren
 ldi wl,0b111               ;Relais
 out portb,wl               ;erstmal aus
 ldi wl,0b011               ;Relais als Ausgang,
 out ddrb,wl                ;Int0 als Eingang
 ldi wl,(1<<7)|(1<<6)       ;Digital-Portlogik der ADC-Ports
 out didr0,wl               ;deaktivieren
 ldi wl,1<<acd              ;Analog-Comparator
 out acsr,wl                ;deaktivieren
 ldi wl,1<<se               ;Sleepmode "Idle" und Low-Level-Int
 out mcucr,wl               ;vorwhlen
 ldi wl,1<<int0             ;ext. Interrupt
 out gimsk,wl               ;freigeben
 rcall lcd_init             ;LCD initialisieren
 ldi anz,anz0*2             ;Termin fr erste Messwertanzeige etwas hher
 ;erste ADC-Messung initialisieren
 ldi wl,(1<<aden)|(1<<adsc)|(1<<adate)|(1<<adie)+3  ;ADC mit Vorteiler 1:8
 out adcsra,wl              ;FreeRun mit Interrupt einschalten
 ldi mb,0                   ;mit Messbereich 0 (8V) 
 rcall sub_neumb            ;beginnen
 sts cindex,yh              ;Scrolltext ab Position 0
 sei                        ;Interrupts freigeben
 rcall matilde              ;Sonderzeichen erlernen...

mainloop:           ;Hauptschleife
 tst anz                    ;Anzeige-Zhler abgelaufen?
 breq job_neuerwert         ;ja, Messwerte aufarbeiten und anzeigen...
 sbrc vflags,upsync         ;neuer LCD-Takt da? - nein...
 rjmp job_zeichenausgabe    ;ja, Zeichen ausgeben...
 sleep                      ;Prozessortakt bis zum nchsten Interrupt aus
 rjmp mainloop              ;nochmal...


job_neuerwert:      ;Job der Mainloop, Messwert prfen, umrechnen und anzeigen
 ldi anz,anz0               ;Termin fr nchste Anzeige setzen
 print 4                    ;Bildschirm lschen
 
 ;Summe der Messwerte bilden
 ldi yl,atab                ;Pointer platzieren
 ldd mw0,y+mwn+pl           ;Nachkommawert Plus,
 ldd mw1,y+mwl+pl           ;L-Wert Plus und
 ldd mw2,y+mwh+pl           ;H-Wert Plus holen
 ldd wl,y+mwn+mi            ;Nachkommawert Minus holen
 add mw0,wl                 ;addieren
 ldd wl,y+mwl+mi            ;L-Wert Minus holen,
 adc mw1,wl                 ;mit bertrag addieren
 ldd wl,y+mwh+mi            ;H-Wert Minus holen,
 adc mw2,wl                 ;mit bertrag addieren
 mov wh,mw0                 ;unkalibrierte Summe 
 mov suml,mw1               ;fr spter 
 mov sumh,mw2               ;sichern

 ldi tmp,kalid              ;EEP-Adresse fr Kalibrierwert DC holen
 rcall sub_kali             ;Messwert kalibrieren
 std y+mwn+dc,mw0           ;fr DC-Anzeige kalibrierte
 std y+mwl+dc,mw1           ;Messwerte
 std y+mwh+dc,mw2           ;sichern

 mov mw0,wh                 ;unkalibrierte
 mov mw1,suml               ;Summe
 mov mw2,sumh               ;zurckholen
 ldi tmp,kalia              ;EEP-Adresse fr Kalibrierwert AC holen
 rcall sub_kali             ;Messwert kalibrieren
 std y+mwn+ac,mw0           ;fr AC-Anzeige kalibrierte
 std y+mwl+ac,mw1           ;Messwerte
 std y+mwh+ac,mw2           ;sichern
 ;in atab liegen nun an Position (H-Byte, L-Byte, Nachkomma-Byte):
 ;0,4,8:    kalibrierter DC-Messwert 
 ;1,5,9:    Mittelwert Plus-Eingang
 ;2,6,10:   kalibrierter AC-Messwert
 ;3,7,11:   Mittelwert Minus-Eingang

 ;Ermittlung des Vorzeichens
 ldi yl,atab                ;Pointer platzieren
 ldd wl,y+mwl+pl            ;L-Wert Plus
 ldd tmp,y+mwl+mi           ;L-Wert Minus
 sub wl,tmp                 ;mit bertrag subtrahieren
 ldd wh,y+mwh+pl            ;H-Wert Plus
 ldd tmp,y+mwh+mi           ;H-Wert Minus
 sbc wh,tmp                 ;mit bertrag subtrahieren
 ldi tmp,'+'                ;erstmal positiv vermuten
 brpl vz_plus               ;positiv? - ja, weiter...
 ldi tmp,'-'                ;nein, negativ,
 com wl                     ;Wert umkehren,
 com wh                     ;also wieder
 subi wl,low(-1)            ;positiv
 sbci wh,high(-1)           ;machen
vz_plus:
 subi wl,low(maxdiff)       ;erlaubtes Limit
 sbci wh,high(maxdiff)      ;abziehen
 brsh vz_1                  ;Differenz grer? - ja, Gleichspannung...
 ldi tmp,' '                ;nein, Wechselspannung, Leerzeichen
vz_1:
 sts signum,tmp             ;Polarittszeichen sichern

 ;kalibrierte und skalierte Summe beider Werte ausgeben:
 rcall lcd_data             ;Polarittszeichen ausgeben...

 ldi tmp,' '                ;Leerzeichen
 lds wl,atab+min+pl         ;Minimalwert Plus holen
 cpi wl,0                   ;0 erreicht?
 brne vz_2                  ;nein...
 lds wl,atab+min+mi         ;Minimalwert Minus holen
 cpi wl,0                   ;0 erreicht?
 brne vz_2                  ;nein...
 ldi tmp,acz                ;ja, Leerzeichen durch AC-Zeichen ersetzen
vz_2:
 rcall lcd_data             ;Wechselspannungszeichen ausgeben...

 ldi yl,atab+dc             ;erstmal fr DC kalibrierte Summe adressieren
 lds tmp,signum             ;Vorzeichen holen
 cpi tmp,' '                ;Wechselspannung?
 brne acdc                  ;nein...
 ldi yl,atab+ac             ;fr AC kalibrierte Summe adressieren
acdc:
 sbr vflags,1<<lsu          ;letzte Stelle deaktivieren
 rcall sub_skalier          ;Messwert skalieren und ausgeben
 cbr vflags,1<<lsu          ;letzte Stelle wieder freigeben
 print ' '                  ;Leerzeichen
 print 'V'                  ;Einheitszeichen

#ifdef servicemode  ;nur bei LCD 16x2
 ;Messbereich im Klartext ausgeben:
 clr wh                     ;oberes Messbereichs-Byte erstmal 0
 ldi wl,2                   ;aus 2V wird dann 8V, 32V, 128V oder 512V
 mov tmp,mb                 ;Kopie vom Messbereich-Index als Schleifenzhler
 inc tmp                    ;1 mehr, um auf 0 testen zu knnen
mbprint:
 lsl wl                     ;mal 4,
 rol wh                     ;da alle Messbereiche
 lsl wl                     ;im Verhltnis 1 zu 4
 rol wh                     ;abgestuft sind
 dec tmp                    ;Schleifenzhler runter
 brne mbprint               ;schon 0? - nein, nochmal
 print16 wh,wl,0,using      ;ja, Messbereich ausgeben


 ;Debug, ADC-Test, Plus-Eingang
 locate 0,17                ;zweite Zeile
 sbr vflags,1<<lsu          ;letzte Stelle deaktivieren
 print '+'
 ldi yl,atab+pl             ;Mittelwert Plus-Eingang
 rcall sub_skalier          ;Messwert skalieren und ausgeben
 print ' '

 ;Debug, ADC-Test, Minus-Eingang:
 print '-'
 ldi yl,atab+mi             ;Mittelwert Minus-Eingang
 rcall sub_skalier          ;Messwert skalieren und ausgeben
 cbr vflags,1<<lsu          ;letzte Stelle wieder freigeben
#endif 

 ;Test, ob Messbereichsumschaltung erforderlich ist
 cli
 ldi wl,255                 ;Minimumwert
 sts atab+min+pl,wl         ;Plus und
 sts atab+min+mi,wl         ;Minus entwerten
 lds wl,maxwert             ;bisherigen Spitzenwert holen
 sts maxwert,yh             ;und entwerten
 sei
 cpi wl,255                 ;Messwert ber Limit?
 brne nix_error             ;nein...
 blink 0,0,2                ;ja, Vorzeichen blinken lassen
nix_error:
 cpi mb,3                   ;Messbereich schon ganz oben?
 breq nix_mbplus            ;ja, Umschaltung unterdrcken
 cpi wl,255                 ;max. ADC-Wert erreicht?
 brne nix_mbplus            ;nein, nicht hochschalten...
 inc mb                     ;ja, Index fr Messbereich erhhen
 rcall sub_neumb            ;Relais umschalten...
 rjmp mainloop              ;Job beenden...

nix_mbplus:
 cpi wl,limu                ;unteres Spitzenwert-Limit unterschritten?
 brsh nix_mbminus           ;nein, nicht runterschalten...
 dec mb                     ;ja, Index fr Messbereich vermindern

 cpi mb,255                 ;Unterlauf?
 brne nix_mbplus1           ;nein...
 clr mb                     ;ja, korrigieren
nix_mbplus1:
 rcall sub_neumb            ;Relais umschalten...
nix_mbminus:
 
 ;bei Messwert (fast) 0 den Scrolltext mit Copyright ausgeben:
 cpi wl,limc                ;Referenz unterschritten?
 brsh nix_copy              ;nein, Zahlen stehenlassen...
 subi ctwait,1              ;Wartezeit runter
 sbrc ctwait,7              ;Unterlauf? - nein, Zahlen stehen lassen...
 rcall printcopy            ;ja, Zahlen durch Scrolltext berschreiben...
nix_copy1:
 rjmp mainloop              ;fertig, zurck zur Mainloop...
nix_copy:
 ldi ctwait,ctw0            ;Wartezeit auf Startwert
 sts cindex,yh              ;den nchsten Scrolltext von vorn beginnen
 rjmp mainloop              ;fertig, zurck zur Mainloop...


job_zeichenausgabe: ;Job der Mainloop, gibt ein Zeichen an das LCD aus
 cbr vflags,15              ;Auftrag mit Vorteiler und Trennbit lschen
 rcall lcd_update           ;Zeichen ausgeben...
 rjmp mainloop              ;fertig, zurck zur Mainloop...


sub_skalier:        ;Unterprogramm, skaliert Messwert gem Messbereich
 ;Messwert wird mit 8, 32, 128 oder 512 multipliziert
 ldd mw0,y+mwn              ;Nachkommawert,
 ldd mw1,y+mwl              ;L-Wert und
 ldd mw2,y+mwh              ;H-Wert holen
 mov tmp,mb                 ;Messbereichs-Index als EEP-Adresse fr Kalib-Wert
 rcall sub_kali             ;Messwert gem Messbereich kalibrieren
 clr mw3                    ;MSB lschen
 mov tmp,mb                 ;Messbereich
 lsl tmp                    ;mal 2 plus 3 als Exponent zur Basis 2
 subi tmp,-3                ;fr Skalierung des Messwertes
sub_skalier1:
 lsl mw0                    ;Nachkommawert,
 rol mw1                    ;L-Byte,
 rol mw2                    ;H-Byte und
 rol mw3                    ;Byte3 vom Messwert mit 2 multiplizieren
 dec tmp                    ;Exponent runter
 brne sub_skalier1          ;unten? - nein, nochmal...
 print24 mw3,mw2,mw1,5,using    ;Nesswert gleich ausgeben
 ret                        ;fertig, zurck...


sub_kali:           ;Unterprogramm, kalibriert unskalierte Messwerte
 out eearl,tmp              ;EEP-Adresse fr Kalibrierwert einstellen
 sbi eecr,eere              ;Leseimpuls generieren
 in tmp,eedr                ;Kalibrierwert holen
 tst tmp                    ;wie sieht denn der Kalibrierwert aus?
 brmi sub_kalineg           ;Kalibrierwert negativ? - ja, verkleinern...
 breq sub_kalipos1          ;Kalibrierwert 0? - ja, abbrechen...
sub_kalipos:                ;nein, positiv, vergrern
 add mw0,mw1                ;1/256
 adc mw1,mw2                ;des Inhaltes
 adc mw2,yh                 ;dazulegen
 dec tmp                    ;wie oft?
 brne sub_kalipos           ;nochmal...
sub_kalipos1:
 ret                        ;fertig, zurck
sub_kalineg:
 neg tmp                    ;Wert positiv machen
sub_kalineg1:
 sub mw0,mw1                ;1/256
 sbc mw1,mw2                ;des Inhaltes
 sbc mw2,yh                 ;herausnehmen
 dec tmp                    ;wie oft?
 brne sub_kalineg1          ;nochmal...
 ret                        ;fertig, zurck...


printcopy:          ;Ausgabe Copyright
 push zl                    ;Variablen beschaffen
 push zh
 push xl
 push xh
 push wl
 ldi anz,anzs               ;Anzeige-Intervall flschen
 ldi zl,low(txt_copy*2)     ;Textstring
 ldi zh,high(txt_copy*2)
 ldi xl,low(vlcd)           ;Bildschirm-RAM
 ldi xh,high(vlcd)
 lds tmp,cindex             ;Position
 inc tmp                    ;im
 andi tmp,63                ;String
 sts cindex,tmp             ;erhhen
 lpm r0,z+                  ;Schlssel holen,
 add zl,tmp                 ;Zeichenposition
 adc zh,yh                  ;addieren
printc1:    ;Do
 lpm wl,z+                  ;Stringzeichen holen,
 eor wl,r0                  ;entcrypten,
 st x+,wl                   ;ja, in Bildschirm-RAM legen,
 cpi zl,low(txt_copy*2+66)  ;Stringende im Flash erreicht?
 brne printc2               ;nein...
 ldi zl,low(txt_copy*2+1)   ;Pointer auf Anfang
 ldi zh,high(txt_copy*2+1)  ;vom Textstring
printc2:
 cpi xl,low(vlcd+zln*spn)   ;Bildschirm-RAM voll?
 brne printc1               ;n, nochmal...
 pop wl                     ;Variablen entsorgen
 pop xh
 pop xl
 pop zh
 pop zl
 ret                        ;feddisch, retour...


sub_neumb:          ;Messbereichsumschaltung
 push xh
 push xl
 ldi yl,reltab              ;Pointer auf
 add yl,mb                  ;Relais-Bitmuster,
 ld xl,y                    ;Bitmuster holen
 out portb,xl               ;und schalten
 sbr vflags,1<<mbu          ;Messbereichsnderung anmelden
 clr kaz                    ;neuen Messzyklus beginnen
 ldi anz,anz0u              ;Ausgabetakt synchronisieren
 ldi xl,low(relwait)        ;Wartezeit 
 ldi xh,high(relwait)       ;einstellen
neumb1:             ;auf Schalten der Relais und Einschwingen des OPV warten
 rcall tunix                ;7 Takte vertrdeln
 sbiw xh:xl,1               ;Wartezeit runter
 brne neumb1                ;abgelaufen? - nein, nochmal...
 pop xl
 pop xh
tunix:              ;7 Takte (Sprungdauer) warten
 ret                        ;fertig, zurck...


f_muxtab:       ;Tabelle mit Mess-Quellen der Folgemessung
.db mqpos,mqneg,mqneg,mqpos ;positiv, negativ, negativ, positiv

f_reltab:       ;Tabelle mit Relais-Bitmustern (PullUp, K2, K1)
.db 0b101,0b111,0b110,0b100 ;8V, 32V, 128V, 512V




